@@ -9,14 +9,17 @@
    <hans-olof.hermansson@postnet.se>, Tor Rustad <Tor.Rustad@bbs.no>,
    Kjetil Barvik <kjetil.barvik@bbs.no>, James Sweeny <jsweeny@us.ibm.com>,
    Chris Ridd <chris.ridd@isode.com>, and several other people whose names
-   I've misplaced.  This code grew slowly over time without much design or
-   planning, with features being tacked on as required.  It's not
-   representative of my normal coding style.
+   I've misplaced (a number of those email addresses probably no longer
+   work, since this code has been around for awhile).
 
    Available from http://www.cs.auckland.ac.nz/~pgut001/dumpasn1.c.
-   Last updated 22 February 2003 (version 20030222, if you prefer it that
-   way).  To build under Windows, use 'cl /MD dumpasn1.c'.  To build on OS390
-   or z/OS, use '/bin/c89 -D OS390 -o dumpasn1 dumpasn1.c'.
+   Last updated 7 January 2009 (version 20090107, if you prefer it that
+   way).  To build under Windows, use 'cl /MD dumpasn1.c'.  To build on 
+   OS390 or z/OS, use '/bin/c89 -D OS390 -o dumpasn1 dumpasn1.c'.
+
+   This code grew slowly over time without much design or planning, and with
+   extra features being tacked on as required.  It's not representative of
+   my normal coding style.
 
    This version of dumpasn1 requires a config file dumpasn1.cfg to be present
    in the same location as the program itself or in a standard directory
@@ -49,7 +52,7 @@
 
 /* The update string, printed as part of the help screen */
 
-#define UPDATE_STRING	"22 February 2003"
+#define UPDATE_STRING	"7 January 2009"
 
 /* Useful defines */
 
@@ -107,7 +110,7 @@
    so it's best to default to allow-all rather than deny-all */
 
 #if defined( linux ) || defined( __linux__ ) || defined( sun ) || \
-	defined( __bsdi__ ) || defined( __FreeBSD__ ) || \
+	defined( __bsdi__ ) || defined( __FreeBSD__ ) || defined( __NetBSD__ ) || \
 	defined( __OpenBSD__ ) || defined( __hpux ) || defined( _M_XENIX ) || \
 	defined( __osf__ ) || defined( _AIX ) || defined( __MACH__ )
   #define __UNIX__
@@ -148,6 +151,12 @@
   extern unsigned _stklen = 16384;
 #endif /* __TURBOC__ */
 
+/* Turn off pointless VC++ warnings */
+
+#ifdef _MSC_VER
+  #pragma warning( disable: 4018 )
+#endif /* VC++ */
+
 /* When we dump a nested data object encapsulated within a larger object, the
    length is initially set to a magic value which is adjusted to the actual
    length once we start parsing the object */
@@ -248,6 +257,7 @@ static int reverseBitString = FALSE;/* Natural order on OS390 is the same as ASN
 #endif /* __OS390__ */
 static int rawTimeString = FALSE;	/* Print raw time strings */
 static int shallowIndent = FALSE;	/* Perform shallow indenting */
+static int outputWidth = 80;		/* 80-column display */
 
 /* The indent size and fixed indent string to the left of the data */
 
@@ -259,12 +269,6 @@ static int shallowIndent = FALSE;	/* Perform shallow indenting */
 #define INDENT_STRING	"         : "
 #endif /* 0 */
 
-/* The width of the output window. This isn't very consistently enforced,
-   for example for hex dumps we always dump 16 bytes at a time to give a nice
-   fixed-format display */
-
-#define OUTPUT_WIDTH		80
-
 /* Error and warning information */
 
 static int noErrors = 0;			/* Number of errors found */
@@ -284,8 +288,9 @@ static FILE *output;				/* Output stream */
 
 typedef struct tagOIDINFO {
 	struct tagOIDINFO *next;		/* Next item in list */
-	char oid[ MAX_OID_SIZE ], *comment, *description;
-	int oidLength;					/* Name, rank, serial number */
+	unsigned char oid[ MAX_OID_SIZE ];
+	int oidLength;
+	char *comment, *description;	/* Name, rank, serial number */
 	int warn;						/* Whether to warn if OID encountered */
 	} OIDINFO;
 
@@ -355,8 +360,10 @@ static const char *configPaths[] = {
 	   else uses */
 	"$HOME/BIN/",
   #else
-	/* Debian has specific places where you're supposed to dump things */
-	"$HOME/", "/etc/dumpasn1/",
+	/* Debian has specific places where you're supposed to dump things.  Note
+	   the dot after $HOME, since config files are supposed to start with a
+	   dot for Debian */
+	"$HOME/.", "/etc/dumpasn1/",
   #endif /* DEBIAN-specific paths */
 
 	/* General environment-based paths */
@@ -378,7 +385,7 @@ static const char *configPaths[] = {
 
 /* Return descriptive strings for universal tags */
 
-char *idstr( const int tagID )
+static char *idstr( const int tagID )
 	{
 	switch( tagID )
 		{
@@ -447,11 +454,12 @@ static OIDINFO *getOIDinfo( char *oid, const int oidLength )
 	{
 	OIDINFO *oidPtr;
 
-	memset( oid + oidLength, 0, 2 );
 	for( oidPtr = oidList; oidPtr != NULL; oidPtr = oidPtr->next )
+		{
 		if( oidLength == oidPtr->oidLength - 2 && \
 			!memcmp( oidPtr->oid + 2, oid, oidLength ) )
 			return( oidPtr );
+		}
 
 	return( NULL );
 	}
@@ -469,7 +477,10 @@ static int addAttribute( char **buffer, char *attribute )
 	return( TRUE );
 	}
 
-/* Table to identify valid string chars (taken from cryptlib) */
+/* Table to identify valid string chars (taken from cryptlib).  Note that
+   IA5String also allows control chars, but we warn about these since
+   finding them in a certificate is a sign that there's something
+   seriously wrong */
 
 #define P	1						/* PrintableString */
 #define I	2						/* IA5String */
@@ -556,7 +567,7 @@ static int readLine( FILE *file, char *buffer )
 			break;
 			}
 
-		/* Make sure the line is of the correct length */
+		/* Make sure that the line is of the correct length */
 		if( bufCount > MAX_LINESIZE )
 			{
 			printf( "Config file line %d too long.\n", lineNo );
@@ -572,8 +583,10 @@ static int readLine( FILE *file, char *buffer )
 
 	/* If we've just passed a CR, check for a following LF */
 	if( ch == '\r' )
+		{
 		if( ( ch = getc( file ) ) != '\n' )
 			ungetc( ch, file );
+		}
 
 	/* Skip trailing whitespace and add der terminador */
 	while( bufCount > 0 &&
@@ -583,14 +596,97 @@ static int readLine( FILE *file, char *buffer )
 
 	/* Handle special-case of ^Z if file came off an MSDOS system */
 	if( ch == CPM_EOF )
+		{
 		while( !feof( file ) )
+			{
 			/* Keep going until we hit the true EOF (or some sort of error) */
 			ch = getc( file );
+			}
+		}
 
 	return( ferror( file ) ? FALSE : TRUE );
 	}
 
-/* Process an OID specified as space-separated hex digits */
+/* Process an OID specified as space-separated decimal or hex digits */
+
+static int processOID( OIDINFO *oidInfo, char *string )
+	{
+	unsigned char binaryOID[ MAX_OID_SIZE ];
+	int firstValue, value, valueIndex = 0, oidIndex = 3;
+
+	memset( binaryOID, 0, MAX_OID_SIZE );
+	binaryOID[ 0 ] = OID;
+	while( *string && oidIndex < MAX_OID_SIZE )
+		{
+		if( oidIndex >= MAX_OID_SIZE - 4 )
+			{
+			printf( "Excessively long OID in config file line %d.\n", 
+					lineNo );
+			return( FALSE );
+			}
+		if( sscanf( string, "%d", &value ) != 1 || value < 0 )
+			{
+			printf( "Invalid value in config file line %d.\n", lineNo );
+			return( FALSE );
+			}
+		if( valueIndex == 0 )
+			{
+			firstValue = value;
+			valueIndex++;
+			}
+		else
+			{
+			if( valueIndex == 1 )
+				{
+				if( firstValue < 0 || firstValue > 2 || value < 0 || \
+					( ( firstValue < 2 && value > 39 ) || \
+					  ( firstValue == 2 && value > 175 ) ) )
+					{
+					printf( "Invalid value in config file line %d.\n", 
+							lineNo );
+					return( FALSE );
+					}
+				binaryOID[ 2 ] = ( firstValue * 40 ) + value;
+				valueIndex++;
+				}
+			else
+				{
+				int hasHighBits = FALSE;
+
+				if( value >= 0x200000L )					/* 2^21 */
+					{
+					binaryOID[ oidIndex++ ] = 0x80 | ( value >> 21 );
+					value %= 0x200000L;
+					hasHighBits = TRUE;
+					}
+				if( ( value >= 0x4000 ) || hasHighBits )	/* 2^14 */
+					{
+					binaryOID[ oidIndex++ ] = 0x80 | ( value >> 14 );
+					value %= 0x4000;
+					hasHighBits = TRUE;
+					}
+				if( ( value >= 0x80 ) || hasHighBits )		/* 2^7 */
+					{
+					binaryOID[ oidIndex++ ] = 0x80 | ( value >> 7 );
+					value %= 128;
+					}
+				binaryOID[ oidIndex++ ] = value;
+				}
+			}
+		while( *string && isdigit( *string ) )
+			string++;
+		if( *string && *string++ != ' ' )
+			{
+			printf( "Invalid OID string in config file line %d.\n", lineNo );
+			return( FALSE );
+			}
+		}
+	binaryOID[ 1 ] = oidIndex - 2;
+	memcpy( oidInfo->oid, binaryOID, oidIndex );
+	oidInfo->oidLength = oidIndex;
+
+	return( TRUE );
+	}
 
 static int processHexOID( OIDINFO *oidInfo, char *string )
 	{
@@ -625,8 +721,9 @@ static int processHexOID( OIDINFO *oidInfo, char *string )
 
 static int readConfig( const char *path, const int isDefaultConfig )
 	{
-	OIDINFO dummyOID = { NULL, "Dummy", "Dummy", "Dummy", 1 }, *oidPtr;
+	OIDINFO dummyOID = { NULL, "Dummy", 0, "Dummy", "Dummy", 1 }, *oidPtr;
 	FILE *file;
+	int seenHexOID = FALSE;
 	char buffer[ MAX_LINESIZE ];
 	int status;
 
@@ -672,8 +769,8 @@ static int readConfig( const char *path, const int isDefaultConfig )
 		/* Check for an attribute tag */
 		if( !strncmp( buffer, "OID = ", 6 ) )
 			{
-			/* Make sure all the required attributes for the current OID are
-			   present */
+			/* Make sure that all of the required attributes for the current
+			   OID are present */
 			if( oidPtr->description == NULL )
 				{
 				printf( "OID ending on config file line %d has no "
@@ -694,8 +791,17 @@ static int readConfig( const char *path, const int isDefaultConfig )
 			memset( oidPtr, 0, sizeof( OIDINFO ) );
 
 			/* Add the new OID */
-			if( !processHexOID( oidPtr, buffer + 6 ) )
-				return( FALSE );
+			if( !strncmp( buffer + 6, "06", 2 ) )
+				{
+				seenHexOID = TRUE;
+				if( !processHexOID( oidPtr, buffer + 6 ) )
+					return( FALSE );
+				}
+			else
+				{
+				if( !processOID( oidPtr, buffer + 6 ) )
+					return( FALSE );
+				}
 			}
 		else if( !strncmp( buffer, "Description = ", 14 ) )
 			{
@@ -740,6 +846,15 @@ static int readConfig( const char *path, const int isDefaultConfig )
 		}
 	fclose( file );
 
+	/* If we're processing an old-style config file, tell the user to 
+	   upgrade */
+	if( seenHexOID )
+		{
+		puts( "\nWarning: Use of old-style hex OIDs detected in "
+			  "configuration file, please\n         update your dumpasn1 "
+			  "configuration file.\n" );
+		}
+
 	return( status );
 	}
 
@@ -781,8 +896,10 @@ static void buildConfigPath( char *path, const char *pathTemplate )
 		else
 			substringSize = pathLen - pathPos;
 		if( substringSize > 0 )
+			{
 			memcpy( newPath + newPathPos, pathBuffer + pathPos,
 					substringSize );
+			}
 		newPathPos += substringSize;
 		pathPos += substringSize;
 
@@ -955,18 +1072,18 @@ static void dumpHex( FILE *inFile, long length, int level, int isInteger )
 	long noBytes = length;
 	int zeroPadded = FALSE, warnPadding = FALSE, warnNegative = isInteger;
 	int singleLine = FALSE;
-	int maxLevel = ( doPure ) ? 15 : 8, i;
+	int maxLevel = ( doPure ) ? 15 : 8, prevCh = -1, i;
 
 	/* Check if LHS status info + indent + "OCTET STRING" string + data will
 	   wrap */
 	if( ( ( doPure ) ? 0 : INDENT_SIZE ) + ( level * 2 ) + 12 + \
-		( length * 3 ) < OUTPUT_WIDTH )
+		( length * 3 ) < outputWidth )
 		singleLine = TRUE;
 
 	if( noBytes > 128 && !printAllData )
 		noBytes = 128;	/* Only output a maximum of 128 bytes */
 	if( level > maxLevel )
-		level = maxLevel;	/* Make sure we don't go off edge of screen */
+		level = maxLevel;	/* Make sure that we don't go off edge of screen */
 	printable[ 8 ] = printable[ 0 ] = '\0';
 	for( i = 0; i < noBytes; i++ )
 		{
@@ -975,7 +1092,7 @@ static void dumpHex( FILE *inFile, long length, int level, int isInteger )
 		if( !( i % lineLength ) )
 			{
 			if( singleLine )
-				putchar( ' ' );
+				fputc( ' ', output );
 			else
 				{
 				if( dumpText )
@@ -998,15 +1115,22 @@ static void dumpHex( FILE *inFile, long length, int level, int isInteger )
 
 		/* If we need to check for negative values and zero padding, check
 		   this now */
-		if( !i )
+		if( i == 0 )
 			{
+			prevCh = ch;
 			if( !ch )
 				zeroPadded = TRUE;
 			if( !( ch & 0x80 ) )
 				warnNegative = FALSE;
 			}
-		if( i == 1 && zeroPadded && ch < 0x80 )
-			warnPadding = TRUE;
+		if( i == 1 )
+			{
+			/* Check for the first 9 bits being identical */
+			if( ( prevCh == 0x00 ) && ( ( ch & 0x80 ) == 0x00 ) )
+				warnPadding = TRUE;
+			if( ( prevCh == 0xFF ) && ( ( ch & 0x80 ) == 0x80 ) )
+				warnPadding = TRUE;
+			}
 		}
 	if( dumpText )
 		{
@@ -1049,6 +1173,45 @@ static void dumpHex( FILE *inFile, long length, int level, int isInteger )
 		}
 	}
 
+/* Convert a binary OID to its string equivalent */
+
+static int oidToString( char *textOID, const unsigned char *oid, 
+						const int oidLength )
+	{
+	int i, j, length;
+	long value;
+
+	/* The first two levels are encoded into one byte since the root level 
+	   has only 3 nodes (40*x + y), however if x = joint-iso-itu-t(2) then 
+	   y may be > 39, so we have to add special-case handling for this */
+	i = oid[ 0 ] / 40;
+	j = oid[ 0 ] % 40;
+	if( i > 2 )
+		{
+		/* Handle special case for large j if i == 2 */
+		j += ( i - 2 ) * 40;
+		i = 2;
+		}
+	length = sprintf( textOID, "%d %d", i, j );
+	for( i = 1, value = 0; i < oidLength; i++ )
+		{
+		const unsigned char data = oid[ i ];
+		const long valTmp = value << 7;
+		int subLen;
+
+		value = valTmp | ( data & 0x7F );
+		if( !( data & 0x80 ) )
+			{
+			subLen = sprintf( textOID + length, " %ld", value );
+			length += subLen;
+			value = 0;
+			}
+		}
+	textOID[ length ] = '\0';
+
+	return( length );
+	}
+
 /* Dump a bitstring, reversing the bits into the standard order in the
    process */
 
@@ -1084,17 +1247,21 @@ static void dumpBitString( FILE *inFile, const int length, const int unused,
 			if( bitString & currentBitMask )
 				value |= bitFlag;
 			if( !( bitString & remainderMask ) )
+				{
 				/* The last valid bit should be a one bit */
 				errorStr = "Spurious zero bits in bitstring";
+				}
 			bitFlag <<= 1;
 			bitString <<= 1;
 			}
 		if( noBits < sizeof( int ) && \
 			( ( remainderMask << noBits ) & value ) )
+			{
 			/* There shouldn't be any bits set after the last valid one.  We
 			   have to do the noBits check to avoid a fencepost error when
 			   there's exactly 32 bits */
 			errorStr = "Spurious one bits in bitstring";
+			}
 		}
 	else
 		value = bitString;
@@ -1164,7 +1331,7 @@ static void displayString( FILE *inFile, long length, int level,
 	if( !doTimeStr && length <= 40 )
 		fprintf( output, " '" );		/* Print string on same line */
 	if( level > maxLevel )
-		level = maxLevel;	/* Make sure we don't go off edge of screen */
+		level = maxLevel;	/* Make sure that we don't go off edge of screen */
 	for( i = 0; i < noBytes; i++ )
 		{
 		int ch;
@@ -1192,10 +1359,9 @@ static void displayString( FILE *inFile, long length, int level,
 			else
 				{
 				const wchar_t wCh = ( ch << 8 ) | getc( inFile );
-#if defined( __WIN32__ ) || ( defined( __UNIX__ ) && !defined( __MACH__ ) )
-				unsigned char outBuf[ 8 ];
-#else
-				char outBuf[ 8 ], *p;
+				char outBuf[ 8 ];
+#ifdef __OS390__
+				char *p;
 #endif /* OS-specific charset handling */
 				int outLen;
 
@@ -1208,21 +1374,41 @@ static void displayString( FILE *inFile, long length, int level,
 				   which the first character looks like a single ASCII char */
 				outLen = wcstombs( outBuf, &wCh, 1 );
 				if( outLen < 1 )
+					{
 					/* Can't be displayed as Unicode, fall back to
 					   displaying it as normal text */
 					ungetc( wCh & 0xFF, inFile );
+					}
 				else
 					{
 					lineLength++;
 					i++;	/* We've read two characters for a wchar_t */
-#if defined( __WIN32__ ) || ( defined( __UNIX__ ) && !defined( __MACH__ ) )
-					wprintf( L"%c", wCh );
+#if defined( __WIN32__ )
+					fputwc( wCh, output );
+#elif defined( __UNIX__ ) && !( defined( __MACH__ ) || defined( __OpenBSD__ ) )
+					/* Some Unix environments differentiate between char
+					   and wide-oriented stdout (!!!), so it's necessary to
+					   manually switch the orientation of stdout to make it
+					   wide-oriented before calling a widechar output
+					   function or nothing will be output (exactly what
+					   level of braindamage it takes to have an
+					   implementation function like this is a mystery).  In
+					   order to safely display widechars, we therefore have
+					   to use the fwide() kludge function to change stdout
+					   modes around the display of the widechar */
+					if( fwide( output, 1 ) > 0 )
+						{
+						fputwc( wCh, output );
+						fwide( output, -1 );
+						}
+					else
+						fputc( wCh, output );
 #else
+  #ifdef __OS390__
 					/* This could use some improvement */
-  #ifndef __MACH__
 					for( p = outBuf; *p != '\0'; p++ )
 						*p = asciiToEbcdic( *p );
-  #endif /* OS X */
+  #endif /* IBM ASCII -> EBCDIC conversion */
 					fprintf( output, "%s", outBuf );
 #endif /* OS-specific charset handling */
 					fPos += 2;
@@ -1246,8 +1432,10 @@ static void displayString( FILE *inFile, long length, int level,
 						ch = '.';	/* Convert non-ASCII to placeholders */
 					}
 				else
+					{
 					if( !isprint( ch ) )
 						ch = '.';	/* Convert non-ASCII to placeholders */
+					}
 #ifdef __OS390__
 				ch = asciiToEbcdic( ch );
 #endif /* __OS390__ */
@@ -1268,8 +1456,10 @@ static void displayString( FILE *inFile, long length, int level,
 
 			case STR_BMP_REVERSED:
 				if( i == noBytes - 1 && ( noBytes & 1 ) )
+					{
 					/* Odd-length BMP string, complain */
 					warnBMP = TRUE;
+					}
 
 				/* Wrong-endianness BMPStrings (Microsoft Unicode) can't be
 				   handled through the usual widechar-handling mechanism
@@ -1316,6 +1506,7 @@ static void displayString( FILE *inFile, long length, int level,
 			}
 		}
 	else
+		{
 		if( doTimeStr )
 			{
 			const char *timeStrPtr = ( strOption == STR_UTCTIME ) ? \
@@ -1334,6 +1525,7 @@ static void displayString( FILE *inFile, long length, int level,
 			}
 		else
 			fputc( '\'', output );
+		}
 	fputc( '\n', output );
 
 	/* Display any problems we encountered */
@@ -1372,7 +1564,7 @@ static long getValue( FILE *inFile, const long length )
 
 /* Get an ASN.1 objects tag and length */
 
-int getItem( FILE *inFile, ASN1_ITEM *item )
+static int getItem( FILE *inFile, ASN1_ITEM *item )
 	{
 	int tag, length, index = 0;
 
@@ -1418,9 +1610,11 @@ int getItem( FILE *inFile, ASN1_ITEM *item )
 
 		length &= LEN_MASK;
 		if( length > 4 )
+			{
 			/* Impossible length value, probably because we've run into
 			   the weeds */
 			return( -1 );
+			}
 		item->headerSize += length;
 		item->length = 0;
 		if( !length )
@@ -1442,7 +1636,7 @@ int getItem( FILE *inFile, ASN1_ITEM *item )
 
 /* Check whether a BIT STRING or OCTET STRING encapsulates another object */
 
-static int checkEncapsulate( FILE *inFile, const int tag, const int length )
+static int checkEncapsulate( FILE *inFile, const int length )
 	{
 	ASN1_ITEM nestedItem;
 	const int currentPos = fPos;
@@ -1458,12 +1652,35 @@ static int checkEncapsulate( FILE *inFile, const int tag, const int length )
 	fPos = currentPos;
 	fseek( inFile, -diffPos, SEEK_CUR );
 
-	/* If it fits exactly within the current item and has a valid-looking
-	   tag, treat it as nested data */
-	if( ( ( nestedItem.id & CLASS_MASK ) == UNIVERSAL || \
-		  ( nestedItem.id & CLASS_MASK ) == CONTEXT ) && \
-		( nestedItem.tag > 0 && nestedItem.tag <= 0x31 ) && \
-		nestedItem.length == length - diffPos )
+	/* If it's not a standard tag class, don't try and dig down into it */
+	if( ( nestedItem.id & CLASS_MASK ) != UNIVERSAL && \
+		( nestedItem.id & CLASS_MASK ) != CONTEXT )
+		return( FALSE );
+
+	/* If it doesn't fit exactly within the current item it's not an 
+	   encapsulated object */
+	if( nestedItem.length != length - diffPos )
+		return( FALSE );
+
+	/* If it doesn't have a valid-looking tag, don't try and go any further */
+	if( nestedItem.tag <= 0 || nestedItem.tag > 0x31 )
+		return( FALSE );
+
+	/* Now things get a bit complicated because it's possible to get some 
+	   (very rare) false positives, for example if a NUMERICSTRING of
+	   exactly the right length is nested within an OCTET STRING, since 
+	   numeric values all look like constructed tags of some kind.  To 
+	   handle this we look for nested constructed items that should really
+	   be primitive */
+	if( ( nestedItem.id & FORM_MASK ) == PRIMITIVE )
+		return( TRUE );
+
+	/* It's constructed, make sure that it's something for which it makes
+	   sense as a constructed object.  At worst this will give some false
+	   negatives for really wierd objects (nested constructed strings inside
+	   OCTET STRINGs), but these should probably never occur anyway */
+	if( nestedItem.tag == SEQUENCE || \
+		nestedItem.tag == SET )
 		return( TRUE );
 
 	return( FALSE );
@@ -1471,7 +1688,7 @@ static int checkEncapsulate( FILE *inFile, const int tag, const int length )
 
 /* Check whether a zero-length item is OK */
 
-int zeroLengthOK( const ASN1_ITEM *item )
+static int zeroLengthOK( const ASN1_ITEM *item )
 	{
 	/* An implicitly-tagged NULL can have a zero length.  An occurrence of this
 	   type of item is almost always an error, however OCSP uses a weird status
@@ -1522,7 +1739,7 @@ int zeroLengthOK( const ASN1_ITEM *item )
 
 /* Check whether the next item looks like text */
 
-static int checkForText( FILE *inFile, const int length )
+static STR_OPTION checkForText( FILE *inFile, const int length )
 	{
 	char buffer[ 16 ];
 	int isBMP = FALSE, isUnicode = FALSE;
@@ -1542,9 +1759,11 @@ static int checkForText( FILE *inFile, const int length )
 		sampleLength = fread( buffer, 1, sampleLength, inFile );
 		fseek( inFile, -sampleLength, SEEK_CUR );
 		for( i = 0; i < sampleLength; i++ )
+			{
 			if( !( isalpha( buffer[ i ] ) || isdigit( buffer[ i ] ) || \
 				   isspace( buffer[ i ] ) ) )
 				return( STR_NONE );
+			}
 		return( STR_IA5 );
 		}
 
@@ -1554,10 +1773,12 @@ static int checkForText( FILE *inFile, const int length )
 	if( isdigit( buffer[ 0 ] ) && ( length == 13 || length == 15 ) && \
 		buffer[ length - 1 ] == 'Z' )
 		{
-		/* It looks like a time string, make sure it really is one */
+		/* It looks like a time string, make sure that it really is one */
 		for( i = 0; i < length - 1; i++ )
+			{
 			if( !isdigit( buffer[ i ] ) )
 				break;
+			}
 		if( i == length - 1 )
 			return( ( length == 13 ) ? STR_UTCTIME : STR_GENERALIZED );
 		}
@@ -1582,20 +1803,24 @@ static int checkForText( FILE *inFile, const int length )
 				   undecided, in which case this comment made a bit more
 				   sense) */
 				if( i < sampleLength - 2 )
+					{
 					/* If the last char(s) are zero but preceding ones
 					   weren't, don't treat it as a BMP string.  This can
 					   happen when storing a null-terminated string if the
 					   implementation gets the length wrong and stores the
 					   null as well */
 					isBMP = TRUE;
+					}
 				continue;
 				}
 			else
+				{
 				/* If we thought we were in a BMPString but we've found a
 				   nonzero byte where there should be a zero, it's neither
 				   an ASCII nor BMP string */
 				if( isBMP )
 					return( STR_NONE );
+				}
 			}
 		else
 			{
@@ -1610,8 +1835,10 @@ static int checkForText( FILE *inFile, const int length )
 				continue;
 				}
 			else
+				{
 				if( isUnicode )
 					return( STR_NONE );
+				}
 			}
 		if( buffer[ i ] < 0x20 || buffer[ i ] > 0x7E )
 			return( STR_NONE );
@@ -1640,8 +1867,8 @@ static void dumpHeader( FILE *inFile, const ASN1_ITEM *item )
 	   data, which means it won't always work on streams */
 	if( extraLen > 0 && doDumpHeader > 1 )
 		{
-		/* Make sure we don't print too much data.  This doesn't work for
-		   indefinite-length data, we don't try and guess the length with
+		/* Make sure that we don't print too much data.  This doesn't work
+		   for indefinite-length data, we don't try and guess the length with
 		   this since it involves picking apart what we're printing */
 		if( extraLen > item->length && !item->indefinite )
 			extraLen = ( int ) item->length;
@@ -1663,7 +1890,8 @@ static void dumpHeader( FILE *inFile, const ASN1_ITEM *item )
 
 /* Print a constructed ASN.1 object */
 
-int printAsn1( FILE *inFile, const int level, long length, const int isIndefinite );
+static int printAsn1( FILE *inFile, const int level, long length,
+					  const int isIndefinite );
 
 static void printConstructed( FILE *inFile, int level, const ASN1_ITEM *item )
 	{
@@ -1693,13 +1921,12 @@ static void printConstructed( FILE *inFile, int level, const ASN1_ITEM *item )
 
 /* Print a single ASN.1 object */
 
-void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
+static void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 	{
 	OIDINFO *oidInfo;
 	STR_OPTION stringType;
 	char buffer[ MAX_OID_SIZE ];
 	long value;
-	int x, y;
 
 	if( ( item->id & CLASS_MASK ) != UNIVERSAL )
 		{
@@ -1788,12 +2015,16 @@ void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 	switch( item->tag )
 		{
 		case BOOLEAN:
-			x = getc( inFile );
-			fprintf( output, " %s\n", x ? "TRUE" : "FALSE" );
-			if( x != 0 && x != 0xFF )
+			{
+			int ch;
+
+			ch = getc( inFile );
+			fprintf( output, " %s\n", ch ? "TRUE" : "FALSE" );
+			if( ch != 0 && ch != 0xFF )
 				complain( "BOOLEAN has non-DER encoding", level );
 			fPos++;
 			break;
+			}
 
 		case INTEGER:
 		case ENUMERATED:
@@ -1809,10 +2040,14 @@ void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			break;
 
 		case BITSTRING:
-			if( ( x = getc( inFile ) ) != 0 )
-				fprintf( output, " %d unused bits", x );
+			{
+			int ch;
+
+			if( ( ch = getc( inFile ) ) != 0 )
+				fprintf( output, " %d unused bit%s",
+						 ch, ( ch != 1 ) ? "s" : "" );
 			fPos++;
-			if( !--item->length && !x )
+			if( !--item->length && !ch )
 				{
 				fputc( '\n', output );
 				complain( "Object has zero length", level );
@@ -1822,13 +2057,14 @@ void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 				{
 				/* It's short enough to be a bit flag, dump it as a sequence
 				   of bits */
-				dumpBitString( inFile, ( int ) item->length, x, level );
+				dumpBitString( inFile, ( int ) item->length, ch, level );
 				break;
 				}
 			/* Drop through to dump it as an octet string */
+			}
 
 		case OCTETSTRING:
-			if( checkEncapsulate( inFile, item->tag, item->length ) )
+			if( checkEncapsulate( inFile, item->length ) )
 				{
 				/* It's something encapsulated inside the string, print it as
 				   a constructed item */
@@ -1854,10 +2090,10 @@ void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			break;
 
 		case OID:
-			/* Hierarchical Object Identifier: The first two levels are
-			   encoded into one byte, since the root level has only 3 nodes
-			   (40*x + y).  However if x = joint-iso-itu-t(2) then y may be
-			   > 39, so we have to add special-case handling for this */
+			{
+			char textOID[ 128 ];
+
+			/* Hierarchical Object Identifier */
 			if( item->length > MAX_OID_SIZE )
 				{
 				fprintf( stderr, "\nError: Object identifier length %ld too "
@@ -1868,10 +2104,15 @@ void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 			fPos += item->length;
 			if( ( oidInfo = getOIDinfo( buffer, ( int ) item->length ) ) != NULL )
 				{
+				int length;
+
+				/* Convert the binary OID to text form */
+				length = oidToString( textOID, buffer, ( int ) item->length );
+
 				/* Check if LHS status info + indent + "OID " string + oid
-				   name will wrap */
+				   name + "(" + oid value + ")" will wrap */
 				if( ( ( doPure ) ? 0 : INDENT_SIZE ) + ( level * 2 ) + 18 + \
-					strlen( oidInfo->description ) >= OUTPUT_WIDTH )
+					strlen( oidInfo->description ) + 2 + length >= outputWidth )
 					{
 					fputc( '\n', output );
 					if( !doPure )
@@ -1880,7 +2121,7 @@ void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 					}
 				else
 					fputc( ' ', output );
-				fprintf( output, "%s\n", oidInfo->description );
+				fprintf( output, "%s (%s)\n", oidInfo->description, textOID );
 
 				/* Display extra comments about the OID if required */
 				if( extraOIDinfo && oidInfo->comment != NULL )
@@ -1899,28 +2140,11 @@ void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 				break;
 				}
 
-			/* Pick apart the OID */
-			x = ( unsigned char ) buffer[ 0 ] / 40;
-			y = ( unsigned char ) buffer[ 0 ] % 40;
-			if( x > 2 )
-				{
-				/* Handle special case for large y if x = 2 */
-				y += ( x - 2 ) * 40;
-				x = 2;
-				}
-			fprintf( output, " '%d %d", x, y );
-			value = 0;
-			for( x = 1; x < item->length; x++ )
-				{
-				value = ( value << 7 ) | ( buffer[ x ] & 0x7F );
-				if( !( buffer[ x ] & 0x80 ) )
-					{
-					fprintf( output, " %ld", value );
-					value = 0;
-					}
-				}
-			fprintf( output, "'\n" );
+			/* Print the OID as a text string */
+			oidToString( textOID, buffer, ( int ) item->length );
+			fprintf( output, " '%s'\n", textOID );
 			break;
+			}
 
 		case EOC:
 		case NULLTAG:
@@ -1969,8 +2193,8 @@ void printASN1object( FILE *inFile, ASN1_ITEM *item, int level )
 
 /* Print a complex ASN.1 object */
 
-int printAsn1( FILE *inFile, const int level, long length,
-			   const int isIndefinite )
+static int printAsn1( FILE *inFile, const int level, long length,
+					  const int isIndefinite )
 	{
 	ASN1_ITEM item;
 	long lastPos = fPos;
@@ -2024,18 +2248,22 @@ int printAsn1( FILE *inFile, const int level, long length,
 				fprintf( output, ( doHexValues ) ? "%04lX %02X NDEF: " :
 						 "%4ld %02X NDEF: ", lastPos, item.id | item.tag );
 			else
+				{
 				if( !seenEOC )
 					fprintf( output, ( doHexValues ) ? "%04lX %02X %4lX: " :
 							 "%4ld %02X %4ld: ", lastPos, item.id | item.tag,
 							 item.length );
+				}
 #else
 			if( item.indefinite )
 				fprintf( output, ( doHexValues ) ? "%04lX NDEF: " :
 						 "%4ld NDEF: ", lastPos );
 			else
+				{
 				if( !seenEOC )
 					fprintf( output, ( doHexValues ) ? "%04lX %4lX: " :
 							 "%4ld %4ld: ", lastPos, item.length );
+				}
 #endif
 			}
 
@@ -2059,6 +2287,7 @@ int printAsn1( FILE *inFile, const int level, long length,
 				return( 0 );
 			}
 		else
+			{
 			if( length <= 0 )
 				{
 				if( length < 0 )
@@ -2066,6 +2295,7 @@ int printAsn1( FILE *inFile, const int level, long length,
 				return( 0 );
 				}
 			else
+				{
 				if( length == 1 )
 					{
 					const int ch = fgetc( inFile );
@@ -2085,11 +2315,18 @@ int printAsn1( FILE *inFile, const int level, long length,
 						return( 1 );
 						}
 					}
+				}
+			}
 		}
 	if( status == -1 )
 		{
+		int i;
+
 		fprintf( stderr, "\nError: Invalid data encountered at position "
-				 "%d.\n", fPos );
+				 "%d:", fPos );
+		for( i = 0; i < item.headerSize; i++ )
+			fprintf( stderr, " %02X", item.header[ i ] );
+		fprintf( stderr, ".\n" );
 		exit( EXIT_FAILURE );
 		}
 
@@ -2106,35 +2343,51 @@ int printAsn1( FILE *inFile, const int level, long length,
 
 /* Show usage and exit */
 
-void usageExit( void )
+static void usageExit( void )
 	{
 	puts( "DumpASN1 - ASN.1 object dump/syntax check program." );
-	puts( "Copyright Peter Gutmann 1997 - 2002.  Last updated " UPDATE_STRING "." );
+	puts( "Copyright Peter Gutmann 1997 - 2008.  Last updated " UPDATE_STRING "." );
 	puts( "" );
+
 	puts( "Usage: dumpasn1 [-acdefhlprstuxz] <file>" );
+	puts( "  Input options:" );
 	puts( "       - = Take input from stdin (some options may not work properly)" );
 	puts( "       -<number> = Start <number> bytes into the file" );
 	puts( "       -- = End of arg list" );
-	puts( "       -a = Print all data in long data blocks, not just the first 128 bytes" );
 	puts( "       -c<file> = Read Object Identifier info from alternate config file" );
 	puts( "            (values will override equivalents in global config file)" );
-	puts( "       -d = Print dots to show column alignment" );
-	puts( "       -e = Don't print encapsulated data inside OCTET/BIT STRINGs" );
+	puts( "" );
+
+	puts( "  Output options:" );
 	puts( "       -f<file> = Dump object at offset -<number> to file (allows data to be" );
 	puts( "            extracted from encapsulating objects)" );
+	puts( "       -w<number> = Set width of output, default = 80 columns" );
+	puts( "" );
+
+	puts( "  Display options:" );
+	puts( "       -a = Print all data in long data blocks, not just the first 128 bytes" );
+	puts( "       -d = Print dots to show column alignment" );
 	puts( "       -h = Hex dump object header (tag+length) before the decoded output" );
 	puts( "       -hh = Same as -h but display more of the object as hex data" );
 	puts( "       -i = Use shallow indenting, for deeply-nested objects" );
 	puts( "       -l = Long format, display extra info about Object Identifiers" );
-	puts( "       -o = Don't check validity of character strings hidden in octet strings" );
 	puts( "       -p = Pure ASN.1 output without encoding information" );
-	puts( "       -r = Print bits in BIT STRING as encoded in reverse order" );
-	puts( "       -s = Syntax check only, don't dump ASN.1 structures" );
 	puts( "       -t = Display text values next to hex dump of data" );
+	puts( "" );
+
+	puts( "  Format options:" );
+	puts( "       -e = Don't print encapsulated data inside OCTET/BIT STRINGs" );
+	puts( "       -r = Print bits in BIT STRING as encoded in reverse order" );
 	puts( "       -u = Don't format UTCTime/GeneralizedTime string data" );
 	puts( "       -x = Display size and offset in hex not decimal" );
+	puts( "" );
+
+	puts( "  Checking options:" );
+	puts( "       -o = Don't check validity of character strings hidden in octet strings" );
+	puts( "       -s = Syntax check only, don't dump ASN.1 structures" );
 	puts( "       -z = Allow zero-length items" );
 	puts( "" );
+
 	puts( "Warnings generated by deprecated OIDs require the use of '-l' to be displayed." );
 	puts( "Program return code is the number of errors found or EXIT_SUCCESS." );
 	exit( EXIT_FAILURE );
@@ -2265,6 +2518,17 @@ int main( int argc, char *argv[] )
 					rawTimeString = TRUE;
 					break;
 
+				case 'W':
+					outputWidth = atoi( argPtr + 1 );
+					if( outputWidth < 40 )
+						{
+						puts( "Invalid output width." );
+						exit( EXIT_FAILURE );
+						}
+					while( argPtr[ 1 ] )
+						argPtr++;	/* Skip rest of arg */
+					break;
+
 				case 'X':
 					doHexValues = TRUE;
 					break;
@@ -2322,8 +2586,8 @@ int main( int argc, char *argv[] )
 		long length;
 		int i, status;
 
-		/* Make sure there's something there, and that it has a definite
-		   length */
+		/* Make sure that there's something there, and that it has a
+		   definite length */
 		status = getItem( inFile, &item );
 		if( status == -1 )
 			{
@@ -2351,6 +2615,28 @@ int main( int argc, char *argv[] )
 		fseek( inFile, offset, SEEK_SET );
 		}
 	printAsn1( inFile, 0, LENGTH_MAGIC, 0 );
+	if( !useStdin && offset == 0 )
+		{
+		unsigned char buffer[ 16 ];
+		long position = ftell( inFile );
+
+		/* If we're dumping a standalone ASN.1 object and there's further
+		   data appended to it, warn the user of its existence.  This is a
+		   bit hit-and-miss since there may or may not be additional EOCs
+		   present, dumpasn1 always stops once it knows that the data should
+		   end (without trying to read any trailing EOCs) because data from
+		   some sources has the EOCs truncated, and most apps know that they
+		   have to stop at min( data_end, EOCs ).  To avoid false positives,
+		   we skip at least 4 EOCs worth of data and if there's still more
+		   present, we complain */
+		fread( buffer, 1, 8, inFile );	/* Skip 4 EOCs */
+		if( !feof( inFile ) )
+			{
+			fprintf( output, "Warning: Further data follows ASN.1 data at "
+					 "position %ld.\n", position );
+			noWarnings++;
+			}
+		}
 	fclose( inFile );
 
 	/* Print a summary of warnings/errors if it's required or appropriate */
